//------------------------------------------------------------------------------
//
//  npch.c
//
// ANTZ - 3D data viewer with realtime real-world interaction based on NPE.
//
// ANTZ is hosted at http://temporalzone.com and NPE at http://neuralphysics.org
//
// Written in 2010-2012 by Shane Saxon - contact@saxondigital.net
//
// Please see main.c for a complete list of additional code contributors
//
// To the extent possible under law, the author(s) have dedicated all copyright 
// and related and neighboring rights to this software to the public domain
// worldwide. This software is distributed without any warranty.
//
// Released under the CC0 license, which is GPL compatible.
//
// You should have received a copy of the CC0 Public Domain Dedication along
// with this software (license file named COPYING.txt). If not, see
// http://creativecommons.org/publicdomain/zero/1.0/
//
//------------------------------------------------------------------------------

#include "npch.h"

#include "../npio.h"
#include "npfile.h"
#include <time.h>
//JJ

static clock_t nextTime;
static int global_sample_rate = 50;   // number of milliseconds between channel changes, rate channel data changes=100/global_sample_rate
char** trackNames = NULL;  // each track has a name
int currentMaxTracks = 0;
float** trackData = NULL;  // each track has a list of float values
int nextAvailableHeaderIndex = 0;
#define MAX_TRACK_DATA 10000
int nextAvailableTrackDataIndex = 0;
#define BUFSIZE 10000 // maximum size of a line read from channel file
char trackReadBuffer[BUFSIZE];  // for fscanf formatting string

int getChannelOffset(char* propertyName) {
	int offset = 0;
	char** trackScanner = trackNames;
	while (*trackScanner) {
		if (strstr(*trackScanner, propertyName) != NULL) {
			return offset;
		}
		offset++;
		trackScanner++;
	}
	return -1;
}

#define LIMIT 100
float *nodePropertyMemoryLocations[LIMIT];
int channel[LIMIT];  // the channel feeding each property
float **nodePropertyMemoryNextLocation = &nodePropertyMemoryLocations[0];
int channelIndex = 0;
#define MAX_NODES 10000
static NPnodePtr nodeList[MAX_NODES];

void cdsInit() {
	int i;
	for (i = 0; i < MAX_NODES; i++) nodeList[i] = 0;
	for (i = 0; i < LIMIT; i++) {
		nodePropertyMemoryLocations[i] = NULL;
		channel[i] = 0;
	}
}



void cdsRegisterNode(NPnodePtr node) {
	printf("Registering node %d\n", node->id);
	if (node->id < MAX_NODES)
	nodeList[node->id] = node;
}

void saveChannel(int channelOffset) {
	channel[channelIndex++] = channelOffset;
	printf("Mapped channelOffset %d\n", channelOffset);
}

void cdsMap(int nodeId, char* propertyName) {
	NPnodePtr node = nodeList[nodeId];
	int channelOffset = getChannelOffset(propertyName);
	if (channelOffset == -1) printf("did not find property %s in the channel\n", propertyName);
	if ((node != NULL) && (channelOffset != -1)) {
		printf("got node %d, now get address of property %s\n", node->id, propertyName);
		if (strcmp(propertyName, "translate_x") == 0) { *nodePropertyMemoryNextLocation++ = &node->translate.x; saveChannel(channelOffset); }
		if (strcmp(propertyName, "translate_y") == 0) { *nodePropertyMemoryNextLocation++ = &node->translate.y; saveChannel(channelOffset); }
		if (strcmp(propertyName, "translate_z") == 0) { *nodePropertyMemoryNextLocation++ = &node->translate.z; saveChannel(channelOffset); }
		if (strcmp(propertyName, "scale_x") == 0) { *nodePropertyMemoryNextLocation++ = &node->scale.x; saveChannel(channelOffset); }
		if (strcmp(propertyName, "scale_y") == 0) { *nodePropertyMemoryNextLocation++ = &node->scale.y; saveChannel(channelOffset); }
		if (strcmp(propertyName, "scale_z") == 0) { *nodePropertyMemoryNextLocation++ = &node->scale.z; saveChannel(channelOffset); }	
	}
}

void createTracks(int numberTracks) {
	int i = 0;
	float** trackDataScanner;
	trackDataScanner = trackData = (float**)malloc(sizeof(float*)*(numberTracks+1));

	for (i = 0; i < numberTracks; i++) {
		*(trackDataScanner++) = (float*)malloc(sizeof(float*)*MAX_TRACK_DATA);
		*(trackDataScanner) = NULL;
	}
}

void saveHeader(char* header) {
	char* headerCopy = (char*)malloc(strlen(header)+1);
	strcpy(headerCopy, header);
	*(trackNames + nextAvailableHeaderIndex) = headerCopy;
	nextAvailableHeaderIndex++;
	*(trackNames + nextAvailableHeaderIndex) = NULL;
	printf("Just saved header %s, now have %d headers\n", header, nextAvailableHeaderIndex);
	// If we've reached limit, need to re-allocate both lists
}

void saveHeaders(const char* filePath, char** headers) {
	// NULL terminated list of pointers
	// do we need to associate these with source file?  (maybe if we record tracks..)
	while (*headers != NULL) {
		saveHeader(*headers++);
	}
}

// saves data for each track
bool saveData(char** data) {
	int i = 0;
	float* singleTrack = NULL;
	if (data == NULL) return false;
	for (i = 0; i < nextAvailableHeaderIndex; i++) {
		singleTrack = *(trackData + i);
		*(singleTrack + nextAvailableTrackDataIndex) = (float)atof(*data++);
	}
	printf("Just loaded track %d\n", nextAvailableTrackDataIndex);
	nextAvailableTrackDataIndex++;
	return true;
}

#define BUFSIZE 10000
char tempBuffer[BUFSIZE];
#define MAXTRACKS 1000
char* tempTracks[MAXTRACKS];
char** chData(FILE* fp) {
	int tempTrackIdx = 0;
	char* bufferScan = &tempBuffer[0];

	if (fgets(&tempBuffer[0], BUFSIZE, fp) == NULL) {
		return NULL;
	}
	tempBuffer[strlen(tempBuffer) - 1] = 0;
	tempTracks[tempTrackIdx++] = bufferScan;
	tempTracks[tempTrackIdx] = NULL;
	while (*bufferScan) {
		if (*bufferScan == ',') {
			*bufferScan = 0;
			tempTracks[tempTrackIdx++] = bufferScan+1;
			tempTracks[tempTrackIdx] = NULL;
		} 
		bufferScan++;
	}
	return &tempTracks[0];
}

/////////////////////////////////
// end of support functions
/////////////////////////////////

//init variables, allocate RAM
#define MAX_TRACKS 1000
//------------------------------------------------------------------------------
void npInitCh (void* dataRef)
{
	currentMaxTracks = MAX_TRACKS;
	trackNames = (char**)malloc((currentMaxTracks+1) * sizeof(char*));  // +1 for NULL pointer indicating end of list
	cdsInit();
}

//------------------------------------------------------------------------------
void npCloseCh (void* dataRef)		//clean-up
{
	return;
}


//open the file and read contents into a buffer
//determine the file type, and call either npChMapImport() or npChDataImport()
//------------------------------------------------------------------------------
void npChFileOpen (const char* filePath, void* dataRef)
{
	int size = 0;
	char path[2048];
	char* fullFilePath = NULL;
	FILE* fp = NULL;
	
	npGetCWD (path, &size);
	fullFilePath = strcat(path, filePath);
	fp = fopen(fullFilePath, "r");
	if (fp != NULL) {
		saveHeaders(filePath, chData(fp));
		createTracks(nextAvailableHeaderIndex);
		while (saveData(chData(fp)));
		fclose(fp);
	} else {
		printf("Failed opening %s\n", fullFilePath);
	}
}

//passed in a buffer containing the entire contents of the antzchmap000x.csv
//parses the CSV formatted contents and builds the channel map list
//merging by using multiple imports allowed
//------------------------------------------------------------------------------
void npChMapImport (char* buffer, int size, void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npChDataImport (char* buffer, int size, void* dataRef)	//antzch0001.csv
{
	return;
}

//exports entire channel map to the buffer in CSV format
//the function that calls this is responsible for freeing the buffer when done
//typically would be written to a CSV file or DB then the buffer is freed
//------------------------------------------------------------------------------
void npChMapExport (char* buffer, int *size, void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npChDataExport (char* buffer, int *size, void* dataRef)
{
	return;
}

//all functions can be used to read/write to or from a CSV file or DB
//separate out the buffer to make this possible...
//add or remove ch attribute track pairs
//------------------------------------------------------------------------------
void npChMapAttributeAdd (int channel, int attribute, int track, void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npChMapAttributeDelete (int channel, int attribute, void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npNewChannel (int channel, void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npDeleteChannel (int channel, void* dataRef)
{
	return;
}

								//will remove '.x' from node->channel.x debug zz
//the node->channe.x specifies the channel map to use for attribute mapping
//of the track-attrib pairs added to the ptr-ptr list
//chID = 0 is the null channel, results in no channel mapping
//------------------------------------------------------------------------------
void npChSubscribeNode (NPnodePtr node, void* dataRef)
{
	pData data = (pData) dataRef;
	data->io.ch.channelIndex = -1;
	data->io.ch.channelSize = nextAvailableTrackDataIndex;  // all tracks are same size
	cdsRegisterNode(node);
	cdsMap(node->id, "translate_x");
	cdsMap(node->id, "translate_y");	
	cdsMap(node->id, "translate_z");
	cdsMap(node->id, "scale_x");
	cdsMap(node->id, "scale_y");
	cdsMap(node->id, "scale_z");
	printf("just subscribed node %d\n", node->id);
	return;
}

bool doUpdate = false;
void npUpdateNodes(void *dataRef) {
	pData data = (pData) dataRef;
	float **nodePropertyMemoryScanner =  &nodePropertyMemoryLocations[0];
	float *track = NULL;
	int channelScanIdx = 0;
	if (doUpdate) {
		doUpdate = false;
		while (nodePropertyMemoryScanner != nodePropertyMemoryNextLocation) {
			track = *(trackData + channel[channelScanIdx]);
			**nodePropertyMemoryScanner++ = *(track + data->io.ch.channelIndex);
			channelScanIdx++;
		}
	}
}

//removes any pointers in the list associated with the node passed in
//used when changing channels which first removes the node, 
//then changes the chID and calls npChSubscribeNode to setup the new channel ptrs
//also called by npNodeDelete()
//------------------------------------------------------------------------------
void npChRemoveNode (NPnodePtr node, void* dataRef)
{
	return;
}


//(un)freeze a node in the list (or should we just add and del..) debug zz

//typically called once per cycle
//1st, update new input stream data, copy the last cycle into the track buffers
//2nd, update/calculate logical tracks, they may use raw tracks as inputs
//3rd, update nodes using track to node-attrib ptr lists
//lastly, output streams, such as live IO devices buffers or to a file
//------------------------------------------------------------------------------
void npUpdateCh (void* dataRef)
{
	pData data = (pData) dataRef;
	clock_t lastTime = clock();

	npUpdateNodes(data);

	if (lastTime > nextTime) {
		doUpdate = true;
		nextTime += global_sample_rate;

		//update the current channelIndex, wrap the buffer back to 0
		data->io.ch.channelIndex++;
		if (data->io.ch.channelIndex >= data->io.ch.channelSize)
			data->io.ch.channelIndex = 0;
	}
}


//--- will add some stream functions
//(multiple types DB, CSV, Network UDP & IP, LiveIO - audio, video, TNG3B, mouse...)
//updated with npUpdateCh() 
//------------------------------------------------------------------------------
void npChOpenInputStream (int needToFigureOut, void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npChCloseInputStream (int needToFigureOut, void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npChOpenOutputStream (int needToFigureOut, void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npChCloseOutStream (int needToFigureOut, void* dataRef)
{
	return;
}

/*
//JJ

//init variables, allocate RAM
//------------------------------------------------------------------------------
void npInitCh (void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npCloseCh (void* dataRef)		//clean-up
{
	return;
}

//open the file and read contents into a buffer
//determine the file type, and call either npChMapImport() or npChDataImport()
//------------------------------------------------------------------------------
void npChFileOpen (const char* filePath, void* dataRef)
{
	return;
}

//passed in a buffer containing the entire contents of the antzchmap000x.csv
//parses the CSV formatted contents and builds the channel map list
//merging by using multiple imports allowed
//------------------------------------------------------------------------------
void npChMapImport (char* buffer, int size, void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npChDataImport (char* buffer, int size, void* dataRef)	//antzch0001.csv
{
	return;
}

//exports entire channel map to the buffer in CSV format
//the function that calls this is responsible for freeing the buffer when done
//typically would be written to a CSV file or DB then the buffer is freed
//------------------------------------------------------------------------------
void npChMapExport (char* buffer, int *size, void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npChDataExport (char* buffer, int *size, void* dataRef)
{
	return;
}

//all functions can be used to read/write to or from a CSV file or DB
//separate out the buffer to make this possible...
//add or remove ch attribute track pairs
//------------------------------------------------------------------------------
void npChMapAttributeAdd (int channel, int attribute, int track, void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npChMapAttributeDelete (int channel, int attribute, void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npNewChannel (int channel, void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npDeleteChannel (int channel, void* dataRef)
{
	return;
}

								//will remove '.x' from node->channel.x debug zz
//the node->channe.x specifies the channel map to use for attribute mapping
//of the track-attrib pairs added to the ptr-ptr list
//chID = 0 is the null channel, results in no channel mapping
//------------------------------------------------------------------------------
void npChSubscribeNode (NPnodePtr node, void* dataRef)
{
	return;
}

//removes any pointers in the list associated with the node passed in
//------------------------------------------------------------------------------
void npChRemoveNode (NPnodePtr node, void* dataRef)
{
	//used when changing channels which first removes the node, 
	//then changes the chID and calls npChSubscribeNode to setup the new channel ptrs
	//also called by npNodeDelete()

	//simply exit if the node is not in the list

	return;
}


//(un)freeze a node in the list (or should we just add and del..) debug zz

//typically called once per cycle
//1st, update new input stream data, copy the last cycle into the track buffers
//2nd, update/calculate logical tracks, they may use raw tracks as inputs
//3rd, update nodes using track to node-attrib ptr lists
//lastly, output streams, such as live IO devices buffers or to a file
//------------------------------------------------------------------------------
void npUpdateCh (void* dataRef)							//add to ctrl loop, debug zz
{
	return;
}


//--- will add some stream functions
//(multiple types DB, CSV, Network UDP & IP, LiveIO - audio, video, TNG3B, mouse...)
//updated with npUpdateCh() 
//------------------------------------------------------------------------------
void npChOpenInputStream (int needToFigureOut, void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npChCloseInputStream (int needToFigureOut, void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npChOpenOutputStream (int needToFigureOut, void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npChCloseOutStream (int needToFigureOut, void* dataRef)
{
	return;
}
*/

